# -*- coding: utf-8 -*-

"""Top-level package for Python reflink."""

__author__ = """Ruben De Smet"""
__email__ = 'pypi@rubdos.be'
__version__ = '0.1.3'

from .error import ReflinkImpossibleError
from .reflink import reflink
from .reflink import supported_at



# -*- coding: utf-8 -*-

"""Main module."""

import os

from . import backend


def reflink(oldpath, newpath):
    """ Create a reflink from ``newpath`` to ``oldpath``.
    Raises a ``NotImplementedError`` when the OS does not implement reflink.
    Raises a ``ReflinkImpossibleError`` when trying to reflink between devices,
    Raises an ``IOError`` when:

    * ``oldpath`` is a directory
    * The user has no permissions to create ``newpath``
    * Trying to reflink into a swapfile (ETXTBSY)

    For debugability's sake, ``reflink`` _can_ throw IOError in other cases,
    when the underlying system call fails for another reason.
    However, these reasons cannot be triggered with this basic call.
    See man `IOCTL-FICLONERANGE(2) <http://man7.org/linux/man-pages/man2/ioctl_ficlonerange.2.html>`__
    and ``backend.py`` for more details on these conditions.

    Example code:

        >>> from reflink import reflink
        >>> reflink("large_file.img", "copy_of_file.img") # doctest: +SKIP
        >>>
    """
    # raise NotImplementedError when OS does not implement reflink
    # raise ReflinkImpossibleError when trying to reflink across devices
    backend.clone(oldpath, newpath)


def supported_at(path):
    """ Returns ``True`` when a path on the filesystem supports reflinking,
    ``False`` otherwise.

    Please note that the current implementation verifies this by testing
    whether it's possible to reflink.
    """

    # Currently, there's no systemcall that tests whether a point in the VFS is supported or not.
    # We create a file at ``path``, and clone it.

    a = os.path.join(path, "i_hope_no_one_will_ever_try_this_name.txt")
    b = os.path.join(path, "i_hope_no_one_will_ever_try_that_name.txt")

    with open(a, 'w+') as f:
        f.write(u"abc")
    try:
        reflink(a, b)
        return True
    except:
        pass
        # TODO: there are probably specific exception that we can handle
    finally:
        os.unlink(a)
        if os.path.isfile(b):
            os.unlink(b)

    return False



import sys

from cffi import FFI

REFLINK_ATTR_PRESERVE = 0x1
REFLINK_ATTR_NONE = 0x0

platform = sys.platform
if platform == 'linux2':
    platform = 'linux'

ffibuilder = FFI()
with open("reflink/%s.c" % platform) as source_file:
    ffibuilder.set_source("reflink._backend", source_file.read())

ffibuilder.cdef("""
int reflink_clone_file(char *oldpath, char *newpath);
extern int errno;
""")

if __file__ == '__main__':
    ffibuilder.compile(verbose=True)



import errno
import sys

from ._backend import ffi  # Ignore PyLintBear (E0611)
from ._backend import lib
from .error import ReflinkImpossibleError

if sys.version_info[0] == 3:
    unicode = str

def clone(oldpath, newpath):
    if isinstance(oldpath, unicode):
        oldpath = oldpath.encode(sys.getfilesystemencoding())
    if isinstance(newpath, unicode):
        newpath = newpath.encode(sys.getfilesystemencoding())

    newpath_c = ffi.new('char[]', newpath)
    oldpath_c = ffi.new('char[]', oldpath)

    rc = lib.reflink_clone_file(oldpath_c, newpath_c)

    if rc == 0:
        return  # Success!

    if rc == -1:
        # ioctl returned an error.
        # We investigate in Python and throw
        if lib.errno == errno.EBADF:
            raise NotImplementedError(
                "BUG: errno EBADF when handling ioctl error")
        if lib.errno == errno.EINVAL:
            raise IOError(
                "EINVAL when handling ioctl: The filesystem does not support reflinking the ranges of the given files.")
        if lib.errno == errno.EISDIR:
            raise IOError("Cannot reflink directories")
        if lib.errno == errno.EOPNOTSUPP:
            raise ReflinkImpossibleError("EOPNOTSUPP")
        if lib.errno == errno.EPERM:
            raise IOError("EPERM: no permission to write to %s" % newpath)
        if lib.errno == errno.ETXTBSY:
            raise IOError("ETXTBSY: cannot reflink from/to swapfiles")
        if lib.errno == errno.EXDEV:
            raise ReflinkImpossibleError("EXDEV")
        if lib.errno == errno.ENOENT:
            raise FileNotFoundError(
                "Cannot open source file %s (%i)" % (oldpath, lib.errno))

        raise Exception("BUG: errno %s not implemented" %
                        errno.errorcode[lib.errno])

    if rc == -3:
        raise IOError("Cannot open file %s for writing (%i)" %
                      (newpath, lib.errno))
    if rc == -2:
        if lib.errno == errno.ENOENT:
            raise FileNotFoundError("No such file %s (%i)" %
                                    (oldpath, lib.errno))
        raise IOError("Cannot open source file %s (%i)" % (oldpath, lib.errno))
    if rc == -4:
        raise NotImplementedError()
    if rc == -5:
        raise IOError("Could not copy permissions (errno %s)" %
                      errno.errorcode[lib.errno])
    raise IOError("Unknown exception (errno %s)" % errno.errorcode[lib.errno])
